/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */
package com.portlandwebworks.commons.dao;


import com.portlandwebworks.commons.reflection.FieldUtils;
import org.springframework.beans.BeanUtils;

import javax.persistence.EmbeddedId;
import java.beans.PropertyDescriptor;
import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;


/**
 * Utility class for introspecting entity classes for various DAO-related properties or annotations.  These utility
 * methods were created for internal use but may serve some purpose outside of the innards of this library.
 * 
 * @author Ronnie Northup
 */
public class DaoUtils
{
	/**
	 * Returns true if the entity class contains a named field, and that field is considered "multi-value".  A field is
	 * multi-value if it an instance of a <code>Collection</code> or <code>Map</code>.
	 *
	 * @param entityClass
	 * @param field
	 * @return              <code>true</code> if the named field is multi-value as defined above.
	 * @throws              IllegalArgumentException if the named field does not exist on the entity class or is inaccessible.
	 */
	public static boolean isMultiValueField(Class entityClass, String field)
	{
		// Look at the getter for this field to see if it returns a subclass of Collection.
		PropertyDescriptor desc = BeanUtils.getPropertyDescriptor(entityClass, field);
		if (desc == null) {
			throw new IllegalArgumentException("Field " + field + " does not exist on entity " + entityClass.getName());
		}
		return (Collection.class.isAssignableFrom(desc.getPropertyType()) || Map.class.isAssignableFrom(desc.getPropertyType()));
	}

	/**
	 * Generates a list of all fields that are considered multi-value, as defined in #isMultiValueField.
	 *
	 * @param entityClass
	 *
	 * @return  A <code>List</code> of all field names that are multi-value.
	 */
	public static List<String> getMultiValueFieldNames(Class entityClass) {

		List<String> fieldNames = new ArrayList<String>();

		PropertyDescriptor[] descriptors = BeanUtils.getPropertyDescriptors(entityClass);
		for (PropertyDescriptor descriptor : descriptors) {
			if (Collection.class.isAssignableFrom(descriptor.getPropertyType()) || Map.class.isAssignableFrom(descriptor.getPropertyType())) {
				fieldNames.add(descriptor.getName());
			}
		}
		
		return fieldNames;
	}

	/**
	 * Returns a list of <code>SortSpecification</code> instances representing the combination of all fields on the
	 * entity on which queries should sort/order if no other sort parameters are provided.  Default sort is specified
	 * on an entity using the <code>@DefaultOrderBy</code> annotation.
	 *
	 * @param entityClass
	 * @return
	 */
	public static List<SortSpecification> getDefaultSortSpecifications(Class entityClass)
	{
		List<SortSpecification> sorts = null;

		for (Field field : FieldUtils.getDeclaredFields(entityClass)) {
			DefaultOrderBy dob = field.getAnnotation(DefaultOrderBy.class);
			if (dob != null) {
				if (sorts == null) {
					sorts = new ArrayList<SortSpecification>();
				}
				sorts.add(new SortSpecification(field.getName(), dob.ascending(), dob.priority()));				
			}
		}
		return sorts;
	}

	/**
	 * @param entityClass
	 * @return
	 */
	public static boolean getDefaultSortAscending(Class entityClass) {
		boolean defaultSortAscending = true;


		Field orderByField = getFieldWithAnnotation(entityClass, DefaultOrderBy.class);

		if (orderByField != null) {
			defaultSortAscending = orderByField.getAnnotation(DefaultOrderBy.class).ascending();
		}

		return defaultSortAscending;
	}


	/**
	 * @param entityClass
	 * @return
	 */
	public static Field getFieldWithAnnotation(Class entityClass, Class<? extends Annotation> annotationType) {
		Field annotatedField = null;
		for (Field field : FieldUtils.getDeclaredFields(entityClass)) {
			Annotation annotationToFind = field.getAnnotation(annotationType);
			if (annotationToFind != null) {
				annotatedField = field;
				break;
			}
		}

		return annotatedField;
	}

	public static Field getEmbeddedIdField(Class entityClass) {
		return getFieldWithAnnotation(entityClass, EmbeddedId.class);
	}
}
