/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

package com.portlandwebworks.commons.dao;

/**
 * This software is the property of Portland Webworks.
 *
 * This class works in conjunction with GridDisplayResults to restrict a large result set by returning one page of
 * results at a time.
 *
 * @Author rnorthup
 */
public class GridDisplayOptions
{
	/**
	 * Use this value as page size to return all results, ignoring any page index specifying.
	 */
	public static final int ALL_RESULTS = Integer.MAX_VALUE;

	/**
	 * This static instance of GridDisplayOptions returns all values, effectively disabling the virtual paging
	 * provided by this class.
	 * <br/>
	 * This set of options relies on one field of the returned entity being defined as the default order by field
	 * by using the @DefaultOrderBy annotation. 
	 * <br/>
	 * Note that if the number of search results exceeds Integer#MAX_VALUE, then these
	 * options will actually NOT return all results!
	 *
	 * @see @DefaultOrderBy
	 * @see Integer#MAX_VALUE
	 */
	public static final GridDisplayOptions OPTIONS_SHOW_ALL = new GridDisplayOptions(null, null, 0, ALL_RESULTS);

	private String sortProperty;

	private Boolean ascending = true;

	private int pageIndex;

	private int pageSize;

	public GridDisplayOptions()
	{
	}

	public GridDisplayOptions(int pageIndex, int pageSize)
	{
		this(null, null, pageIndex, pageSize);
	}

	public GridDisplayOptions(String sortProperty, Boolean ascending, int pageIndex, int pageSize)
	{
		this.sortProperty = sortProperty;
		this.ascending = ascending;
		this.pageIndex = pageIndex;
		this.pageSize = pageSize;
	}

	/**
	 * The name of a property on the DAO's entity class to use for sorting the results.
	 *
	 * @return
	 */
	public String getSortProperty()
	{
		return sortProperty;
	}

	public void setSortProperty(String sortProperty)
	{
		this.sortProperty = sortProperty;
	}

	/**
	 * Indicates that results should be sorted in ascending order, instead of descending.  This is the default
	 * sort order if none is specified.
	 *
	 * @return
	 */
	public Boolean isAscending()
	{
		return ascending;
	}

	public void setAscending(Boolean ascending)
	{
		this.ascending = ascending;
	}

	/**
	 * The zero-based page number of results to return from the search.
	 *
	 * @return
	 */
	public int getPageIndex()
	{
		return pageIndex;
	}

	public void setPageIndex(int pageIndex)
	{
		this.pageIndex = pageIndex;
	}

	/**
	 * The number of items that are in a single page of results.
	 * 
	 * @return
	 */
	public int getPageSize()
	{
		return pageSize;
	}

	public void setPageSize(int pageSize)
	{
		this.pageSize = pageSize;
	}
}