/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

package com.portlandwebworks.commons.domain;

import com.portlandwebworks.commons.dao.DefaultOrderBy;

import javax.persistence.*;

/**
 * For lookup tables that represent a set of values coded in another system (for example, as a list of enumerated
 * values in java).  The code field maps to the values in the other system.
 *
 * By default, code value queries order their results by ID.  You can override default sorting by adding
 * a @DefaultOrderBy annotation to your subclass field with a priority value less than 100.
 *
 **/
@MappedSuperclass
public class CodeValueBase<E extends Enum> extends IEntity<Integer> implements Comparable
{
	/**
	 * Code values are not set up with id generators because they will never be added at runtime, since they are tied
	 * to pre-defined enum values in java code.
	 */
	@Id
	@Column(name = "id", precision = 2, scale = 0)
	@DefaultOrderBy(priority = 100)
	private Integer id;

	@Override
	public Integer getId()
	{
		return id;
	}

	@Override
	public void setId(Integer id)
	{
		this.id = id;
	}

	/**
	 * Enum representation of the status value.  WIll be stored in the underlying database table as the text version
	 * of the enum value.
	 */
	@Enumerated(EnumType.STRING)
	@Column(unique = true, nullable = false)
	private E code;

	/**
	 * Human readable text representation of the status value.
	 */
	@Column(length = 100, nullable = false)
	private String name;

	/**
	 * Explanation of that the status value means.
	 */
	@Column(length = 1000)
	private String description;
	

	public CodeValueBase()
	{
	}

	public CodeValueBase(E code, String name)
	{
		this.code = code;
		this.name = name;
	}

	public CodeValueBase(E code, String name, String description)
	{
		this.code = code;
		this.name = name;
		this.description = description;
	}

	/**
	 * Two status values are the same if they share the same type and code.
	 *
	 * @param o
	 * @return
	 */
	@Override
	public boolean equals(Object o)
	{
		if (this == o) return true;
		if (!(o instanceof CodeValueBase)) return false;
		if (!super.equals(o)) return false;

		CodeValueBase that = (CodeValueBase)o;

		if (code != null ? !code.equals(that.code) : that.code != null) return false;

		return true;
	}

	@Override
	public int hashCode()
	{
		int result = super.hashCode();
		result = 31 * result + (code != null ? code.hashCode() : 0);
		return result;
	}

	public E getCode()
	{
		return code;
	}

	public void setCode(E code)
	{
		this.code = code;
	}

	public String getName()
	{
		return name;
	}

	public void setName(String name)
	{
		this.name = name;
	}

	public String getDescription()
	{
		return description;
	}

	public void setDescription(String description)
	{
		this.description = description;
	}

	public int compareTo(Object o)
	{
		if (o.getClass().equals(this.getClass())) {
			return code.compareTo(((CodeValueBase)o).code);
		}
		return 0;
	}
}
