/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

package com.portlandwebworks.commons.email;

import com.portlandwebworks.commons.domain.UserEntityBase;
import com.portlandwebworks.commons.email.template.IEmailTemplate;
import com.portlandwebworks.commons.email.template.IEmailTemplateProvider;
import org.apache.commons.collections.buffer.PriorityBuffer;
import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.builder.ToStringBuilder;
import org.apache.commons.lang.builder.ToStringStyle;
import org.apache.commons.lang.time.DateFormatUtils;
import org.apache.commons.lang.time.DateUtils;
import org.apache.log4j.Logger;
import org.springframework.mail.MailException;
import org.springframework.mail.javamail.JavaMailSenderImpl;
import org.springframework.mail.javamail.MimeMessageHelper;

import javax.mail.Message;
import javax.mail.MessagingException;
import javax.mail.internet.AddressException;
import javax.mail.internet.InternetAddress;
import javax.mail.internet.MimeMessage;
import java.io.UnsupportedEncodingException;
import java.util.*;

/**
 * This software is the property of Portland Webworks.
 *
 * @Author rnorthup
 */
public class EmailSender implements Runnable
{
	private Logger logger = Logger.getLogger(EmailSender.class);

	private Logger queueLogger = Logger.getLogger("EmailSender.MailQueue");

	private IEmailTemplateProvider templateProvider;

	private JavaMailSenderImpl mailSender;

	private String smtpHost;
	private Integer smtpPort;
	private boolean sslEnabled;
	private boolean tlsEnabled;
	private String smtpUser;
	private String smtpPassword;
	private String fromName;
	private String fromEmail;
	private InternetAddress[] ccAddresses;
	private boolean testMode;
	private InternetAddress testAddress;
	private Integer emailSendDelay;
	private boolean suppressDuplicates;
	private static final int DEFAULT_EMAIL_SEND_DELAY = 5000;
	protected InternetAddress senderAddress;
	private InternetAddress replyToAddress;
	private String replyToEmail;
	private String replyToName;

	/**
	 * The component that is responsible for loading and supplying the email sender with templates based on unique name.
	 *
	 * @param templateProvider
	 */
	public void setTemplateProvider(IEmailTemplateProvider templateProvider)
	{
		this.templateProvider = templateProvider;
	}

	/**
	 * Mail server's IP or hostname.
	 * @param smtpHost
	 */
	public void setSmtpHost(String smtpHost)
	{
		this.smtpHost = smtpHost;
	}

	/**
	 * Optionally sets the mail server's listen port.
	 *
	 * @param smtpPort
	 */
	public void setSmtpPort(Integer smtpPort)
	{
		this.smtpPort = smtpPort;
	}

	/**
	 * Enables SSL for communication with the mail server.
	 *
	 * @param sslEnabled
	 */
	public void setSslEnabled(boolean sslEnabled)
	{
		this.sslEnabled = sslEnabled;
	}

	/**
	 * Enabled TLS for communication with the mail server.
	 *
	 * @param tlsEnabled
	 */
	public void setTlsEnabled(boolean tlsEnabled)
	{
		this.tlsEnabled = tlsEnabled;
	}

	/**
	 * The user to impersonate for communication with the mail server.
	 *
	 * @param smtpUser
	 */
	public void setSmtpUser(String smtpUser)
	{
		this.smtpUser = smtpUser;
	}

	/**
	 * The optional password to use for authentication with the mail server.
	 *
	 * @param smtpPassword
	 */
	public void setSmtpPassword(String smtpPassword)
	{
		this.smtpPassword = smtpPassword;
	}

	/**
	 * The name that will appear in the "from" field of outgoing emails.
	 *
	 * @param fromName
	 */
	public void setFromName(String fromName)
	{
		this.fromName = fromName;
	}

	/**
	 * The email address that will appear in the "from" field of outgoing emails.
	 *
	 * @param fromEmail
	 */
	public void setFromEmail(String fromEmail)
	{
		this.fromEmail = fromEmail;
	}

	public String getReplyToEmail()
	{
		return replyToEmail;
	}

	public void setReplyToEmail(String replyToEmail)
	{
		this.replyToEmail = replyToEmail;
	}

	public String getReplyToName()
	{
		return replyToName;
	}

	public void setReplyToName(String replyToName)
	{
		this.replyToName = replyToName;
	}

	/**
	 * The list of email addresses that will be BCC'd on all outgoing emails (for diagnostic purposes of course).
	 *
	 * @param ccAddresses
	 */
	public void setCcAddresses(String ccAddresses)
	{
		try {
			this.ccAddresses = InternetAddress.parse(ccAddresses);
		}
		catch (AddressException e) {
			throw new RuntimeException("Invalid cc address list " + ccAddresses);
		}
	}

	/**
	 * When set to true, the email addresses specified for outgoing emails will be ignored.  This includes the email
	 * addresses in the email field of any UserEntityBase subclass provided.
	 * <p/>
	 * All emails will be sent to the configured testAddress.  If no value is provided for testAddress, no emails will
	 * be sent.
	 *
	 * @param testMode
	 */
	public void setTestMode(boolean testMode)
	{
		this.testMode = testMode;
	}

	/**
	 * The email address to which all emails will be sent if testMode is enabled.
	 *
	 * @param testAddress
	 */
	public void setTestAddress(String testAddress)
	{
		try {
			this.testAddress = new InternetAddress(testAddress);
		}
		catch (AddressException e) {
			throw new RuntimeException("Invalid test email address " + testAddress);
		}
	}

	/**
	 * Sets the number of guaranteed milliseconds that will pass before an email message is sent to the configured SMTP
	 * server.  This allows the sender to cancel a sent message, or to prevent duplicates from being sent.
	 *
	 * @param emailSendDelay
	 */
	public void setEmailSendDelay(int emailSendDelay)
	{
		this.emailSendDelay = emailSendDelay;
	}

	/**
	 * Indicates whether the email sender should check the queue for other emails to the same recipient, with the same
	 * exact email message, and suppress queuing of the email if a duplicate is found.
	 * <p/>
	 * If no #emailSendDelay value has been specified, a default millisecond value of #DEFAULT_EMAIL_SEND_DELAY will be
	 * used.
	 *
	 * @param suppressDuplicates
	 */
	public void setSuppressDuplicates(boolean suppressDuplicates)
	{
		this.suppressDuplicates = suppressDuplicates;

		if (emailSendDelay == null) {
			logger.info("Suppressing duplicates is configured--setting email send delay to default value (" + DEFAULT_EMAIL_SEND_DELAY + " millis)");
			emailSendDelay = DEFAULT_EMAIL_SEND_DELAY;
		}
	}

	/**
	 * @param templateName
	 * @param args
	 * @param users
	 */
	public EmailResult sendUserEmail(String templateName, Map<String, Object> args, Collection<? extends UserEntityBase> users)
	{
		return sendUserEmail(templateName, args, users.toArray(new UserEntityBase[users.size()]));
	}

	public void cancelUserEmail(UserEntityBase user, String templateName, Map<String, Object> args)
	{
		if (emailSendDelay == null) {
			logger.warn("Unable to cancel user email, emailSendDelay is not configured");
			return;
		}

		InternetAddress address = null;
		try {
			address = new InternetAddress(user.getEmail());
			cancelEmail(address, templateName, args);
		}
		catch (AddressException e) {
			logger.warn("Can't cancel email for user with invalid email address " + user.getEmail());
		}
	}

	/**
	 * @param templateName
	 * @param args
	 * @param users
	 */
	public EmailResult sendUserEmail(String templateName, Map<String, Object> args, UserEntityBase... users)
	{
		ArrayList<String> badAddresses = new ArrayList<String>();

		List<InternetAddress> recipients = new ArrayList<InternetAddress>(users.length);
		for (UserEntityBase user : users) {
			try {
				recipients.add(new InternetAddress(user.getEmail(), user.getFirstAndLastName()));
			}
			catch (UnsupportedEncodingException e) {
				logger.warn("Could not build email address from email " + user.getEmail() + ", name " + user.getFirstAndLastName());

				try {
					recipients.add(new InternetAddress(user.getEmail()));
				}
				catch (AddressException e1) {
					badAddresses.add(user.getEmail());
				}
			}
		}
		EmailResult finalResult = sendEmail(templateName, args, recipients.toArray(new InternetAddress[recipients.size()]));

		//TODO: without synchronization or passing a pre-populated result to the send method, there is no way to
		//TODO: guarantee that these bad addresses will get into the result before it is viewed by the caller.
		// Merge in our address parse failures into the final results
		for (String badAddress : badAddresses) {
			finalResult.addBadAddress(badAddress);
		}

		return finalResult;
	}

	/**
	 * @param templateName
	 * @param args
	 * @param recipients
	 */
	public EmailResult sendEmail(String templateName, Map<String, Object> args, Collection<InternetAddress> recipients)
	{
		return sendEmail(templateName, args, recipients.toArray(new InternetAddress[recipients.size()]));
	}

	public void cancelEmail(InternetAddress address, String templateName, Map<String, Object> args)
	{
		synchronized (mailQueue) {
			Iterator iter = mailQueue.iterator();
			while (iter.hasNext()) {
				ScheduledMessage schedMsg = (ScheduledMessage)iter.next();
				if (schedMsg.matches(address, templateName, args)) {
					iter.remove();
				}
			}
		}
	}

	/**
	 * @param templateName
	 * @param args
	 * @param recipients
	 */
	public EmailResult sendEmail(String templateName, Map<String, Object> args, InternetAddress... recipients)
	{
		return sendEmail(templateName, args, null, recipients);
	}

	/**
	 * @param templateName
	 * @param args
	 * @param attachments
	 * @param recipients
	 */
	public EmailResult sendEmail(String templateName, Map<String, Object> args, List<EmailAttachment> attachments, Collection<InternetAddress> recipients)
	{
		return sendEmail(templateName, args, attachments, recipients.toArray(new InternetAddress[recipients.size()]));
	}

	/**
	 * @param templateName
	 * @param args
	 * @param attachments
	 * @param recipients
	 */
	public EmailResult sendEmail(String templateName, Map<String, Object> args, List<EmailAttachment> attachments, InternetAddress... recipients)
	{
		if (mailSender == null) {
			throw new RuntimeException("Mail sender is not initialized.  Did you get to call my initialize() method?");
		}

		EmailResult result = new EmailResult();

		// Make the list of attachments available to the email template.
		if (attachments != null) {
			args.put("attachments", attachments);
		}

		IEmailTemplate template = templateProvider.getTemplate(templateName);

		for (InternetAddress recipient : recipients) {
			MimeMessage msg = null;
			boolean handled = false;
			try {
				template.setParamValues(args);

				msg = mailSender.createMimeMessage();
				ScheduledMessage message = new ScheduledMessage(msg, recipient, templateName, args);
				message.setResult(result);
				result.addPending(msg);

				MimeMessageHelper helper = new MimeMessageHelper(msg, attachments != null, "UTF-8");
				helper.setFrom(senderAddress);

				if (testMode) {
					if (testAddress != null) {
						logger.info("TEST MODE--email recipient has been overridden to " + testAddress.getAddress());
						helper.setTo(testAddress);
					}
					else {
						logger.warn("Test mode enabled, but no test address configured.  Email WILL NOT BE SENT.");
						continue;
					}
				}
				else {
					helper.setTo(recipient);
				}
				helper.setSubject(template.getSubject());
				helper.setText(template.getBody(), template.isHtml());

				if (attachments != null) {

					for (EmailAttachment attachment : attachments) {
						helper.addAttachment(attachment.getFileName(), attachment.getContents());
					}
				}

				if (!ArrayUtils.isEmpty(ccAddresses)) {
					// Need to keep track of every outgoing email so that we can easily track whether emails are being sent as designed.
					for (InternetAddress ccAddress : ccAddresses) {
						msg.addRecipient(Message.RecipientType.BCC, ccAddress);
					}
				}

				// The calling code has indicated that the successful sending of this email is not critical to the
				// operation performed.  Therefore, it is safe to asynchronously queue this message for delivery.
				if (emailSendDelay != null) {
					// This email sender is configured with a minimum millis delay before sending an email.  Set the
					// outgoing email with a time in the future that must be reached before this email is sent to
					// the SMTP server.  Also include the template name and template args so that this email can
					// be cancelled, or so duplicates can be suppressed.
					message.scheduled = DateUtils.addMilliseconds(new Date(), this.emailSendDelay);
					queueMessage(message);
					handled = true;
				}
				else {
					queueMessage(message);
					handled = true;
				}
			}
			catch (MessagingException e) {
				result.setFailure(msg, e);

				logger.error("Unhandled exception building message", e);

				logger.debug("FAILURE: " + e);

				handled = true;
			}
			finally {
				if (! handled) {
					logger.debug("UNKNOWN FAILURE: " + msg);
					
					result.setUnknownFailure(msg);
				}
			}
		}

		return result;
	}

	/**
	 * Internal flag that indicates whether the queue reader should shut itself down.
	 */
	private boolean shouldQuit = false;

	/**
	 * Internal data structure to ensure the proper queuing of outgoing messages based on their "scheduled" date for
	 * delivery.
	 */
	private final PriorityBuffer mailQueue = new PriorityBuffer(true, new Comparator()
	{

		public int compare(Object thisOne, Object thatOne)
		{
			ScheduledMessage thisMessage = (ScheduledMessage)thisOne;
			ScheduledMessage thatMessage = (ScheduledMessage)thatOne;
			if (thisMessage.scheduled != null) {
				if (thatMessage.scheduled != null) {
					// Both messages were scheduled in the future.  Compare their scheduled dates.
					return thisMessage.scheduled.compareTo(thatMessage.scheduled);
				}
				else {
					// Other message has no scheduled date.  That one has higher priority.
					return 1;
				}
			}
			else {
				if (thatMessage.scheduled != null) {
					// Other message was scheduled in the future.  This one has higher priority.
					return -1;
				}
				else {
					// Neither message was scheduled in the future. Neither has higher priority.
					return 0;
				}
			}
		}
	});

	//private final LinkedList<MimeMessage> mailQueue = new LinkedList<MimeMessage>();

	/**
	 * initializer
	 */
	public void initialize()
	{
		if (StringUtils.isBlank(smtpHost)) {
			throw new RuntimeException("smtpHost is required");
		}

		if (fromEmail == null) {
			throw new RuntimeException("fromEmail is required");
		}

		if (tlsEnabled && sslEnabled) {
			throw new RuntimeException("Can't enable both TLS and SSL");
		}

		mailSender = new JavaMailSenderImpl();
		mailSender.setHost(smtpHost);
		if (smtpPort != null) {
			mailSender.setPort(smtpPort);
		}
		mailSender.setUsername(smtpUser);
		if (StringUtils.isNotBlank(smtpPassword)) {
			mailSender.setPassword(smtpPassword);
		}

		Properties javaMailProps = new Properties();
		javaMailProps.put("mail.debug", "true");

		if (sslEnabled) {
			mailSender.setProtocol("smtps");
			javaMailProps.setProperty("mail.smtp.ssl.enable", "true");
		}
		if (tlsEnabled) {
			javaMailProps.setProperty("mail.smtp.starttls.enable", "true");
		}
		if (javaMailProps.size() > 0) {
			mailSender.setJavaMailProperties(javaMailProps);
		}

		senderAddress = null;
		try {
			senderAddress = new InternetAddress(fromEmail, fromName);
		}
		catch (UnsupportedEncodingException e) {
			throw new RuntimeException("Invalid SMTP sender: " + fromName + " <" + fromEmail + ">");
		}

		replyToAddress = null;
		if (replyToEmail != null) {
			try {
				replyToAddress = new InternetAddress(replyToEmail, replyToName);
			}
			catch (UnsupportedEncodingException e) {
				throw new RuntimeException("Invalid SMTP replyTo: " + replyToName + "<" + replyToEmail + ">");
			}
		}

		queueLogger.info("Initializing mail queue...");

		Thread t = new Thread(this, "PWW-Commons Email Queue");
		t.setDaemon(true);
		t.start();
	}

	/**
	 * Awakens the mail queue reader if it is sleeping and tells it to quit.
	 */
	public void shutdown()
	{
		shouldQuit = true;
		synchronized (mailQueue) {
			mailQueue.notify();
		}
	}

	protected void queueMessage(ScheduledMessage message)
	{

		synchronized (mailQueue) {

			if (suppressDuplicates) {
				for (Object obj : mailQueue) {
					if (obj.equals(message)) {
						logger.info("Detected duplicate message " + message + ", queue already contains " + obj);

						EmailResult result = message.result;
						result.setDuplicate(message.message);

						return;
					}
				}
			}

			mailQueue.add(message);
			mailQueue.notify();
		}
	}


	/**
	 * The
	 */
	public void run()
	{

		long waitTime = Long.MAX_VALUE;

		while (!shouldQuit) {

			synchronized (mailQueue) {

				if (mailQueue.isEmpty() || waitTime != Long.MAX_VALUE) {

					if (logger.isInfoEnabled()) {
						if (mailQueue.isEmpty()) {
							logger.info("Mail queue is empty, waiting...");
						}
						else {
							logger.info("Waiting " + waitTime + " millis for first item in mail queue");
						}
					}

					try {
						mailQueue.wait(waitTime);
					}
					catch (InterruptedException e) {
						if (shouldQuit)
							return;
					}

					// Reset the wait time to be indefinite
					waitTime = Long.MAX_VALUE;
				}
			}

			if (!mailQueue.isEmpty()) {

				synchronized (mailQueue) {

					if (!mailQueue.isEmpty()) {
						logger.info("Mail queue is non-empty, size is " + mailQueue.size());

						//MimeMessage nextMessage = mailQueue.removeLast();
						Date now = new Date();
						ScheduledMessage schedMessage = (ScheduledMessage)mailQueue.get();
						if (schedMessage.scheduled != null && schedMessage.scheduled.after(now)) {
							if (logger.isDebugEnabled()) {
								logger.debug("First message in queue is scheduled for future date " + DateFormatUtils.format(schedMessage.scheduled, "MM/dd/yyyy HH:mm:ss"));
							}

							// Highest priority message in the queue is scheduled for some time in the future.  Set this
							// thread's wait time to be some estimated value
							waitTime = schedMessage.scheduled.getTime() - now.getTime();
							continue;
						}

						schedMessage = (ScheduledMessage)mailQueue.remove();

						MimeMessage nextMessage = schedMessage.message;

						EmailResult result = schedMessage.getResult();
						if (result == null) {
							throw new RuntimeException("Message " + schedMessage + " has already been processed");
						}

						if (queueLogger.isInfoEnabled()) {
							try {
								queueLogger.info("Processing queued email: "
										+ new ToStringBuilder(nextMessage).append("from",
										nextMessage.getFrom()).append("to",
										nextMessage.getAllRecipients()[0].toString()).append("subject",
										nextMessage.getSubject()));
							}
							catch (MessagingException e) {
								logger.error("Unhandled exception logging email: " + e.getMessage(), e);
							}
						}

						boolean handled = false;

						try {

							mailSender.send(nextMessage);

							// Move this recipient from the pending list to the success list
							logger.debug("SUCCESS: " + schedMessage.message);

							result.setSuccess(schedMessage.message);

							handled = true;
						}
						catch (MailException e) {
							logger.error("Unhandled exception sending email: " + e.getMessage(), e);

							logger.debug("FAILURE: " + schedMessage.message);

							result.setFailure(schedMessage.message, e);

							handled = true;
						}
						finally {
							// This will ensure that, if there was an unhandled exception, this recipient is pulled from
							// the list of pending recipients and put in the failure bucket.
							if (! handled) {
								logger.debug("UNKNOWN FAILURE: " + schedMessage.message);
								
								result.setUnknownFailure(schedMessage.message);
							}
						}
					}
				}
				logger.info(mailQueue.size() + " emails left in the queue");
			}
		}
	}
}


/**
 * Wrapper class for combining all contextual information about an email message to allow:
 * <ol>
 * <li>Scheduling an email to be sent in the future</li>
 * <li>Cancelling an email based on the nature of the email message</li>
 * </ol>
 */
class ScheduledMessage
{
	public MimeMessage message;

	public Date scheduled;

	public InternetAddress address;

	public String templateName;

	public Map<String, Object> args;

	public EmailResult result;

	/**
	 * This message will be sent immediately.
	 *
	 * @param message
	 */
	ScheduledMessage(MimeMessage message, InternetAddress recipient)
	{
		this.message = message;
		this.address = recipient;
	}

	/**
	 * This message will be sent on or after the scheduled date/time.
	 *
	 * @param message
	 * @param scheduled
	 */
	ScheduledMessage(MimeMessage message, InternetAddress recipient, Date scheduled)
	{
		this.message = message;
		this.address = recipient;
		this.scheduled = scheduled;
	}

	/**
	 *
	 *
	 * @param message
	 * @param templateName
	 * @param args
	 */
	ScheduledMessage(MimeMessage message, InternetAddress recipient, String templateName, Map<String, Object> args)
	{
		this.message = message;
		this.address = recipient;
		this.templateName = templateName;
		this.args = args;
	}

	public EmailResult getResult()
	{
		return result;
	}

	public void setResult(EmailResult result)
	{
		this.result = result;
	}

	/**
	 * A message is equal if it has the same recipient and content (template name, and template arguments).
	 *
	 * @param o
	 * @return
	 */
	@Override
	public boolean equals(Object o)
	{
		if (this == o) return true;
		if (o == null || getClass() != o.getClass()) return false;

		ScheduledMessage that = (ScheduledMessage)o;

		if (address != null ? !address.equals(that.address) : that.address != null) return false;
		if (args != null ? !args.equals(that.args) : that.args != null) return false;
		if (templateName != null ? !templateName.equals(that.templateName) : that.templateName != null) return false;

		return true;
	}

	/**
	 * A message matches another if it has the same recipient and content (template name, and template arguments).  This
	 * method has the same logic as the equals() implementation, but can be applied
	 *
	 * @param address
	 * @param templateName
	 * @param args
	 * @return
	 */
	public boolean matches(InternetAddress address, String templateName, Map<String, Object> args)
	{
		if (this.address != null && !this.address.equals(address)) return false;
		if (this.templateName != null && !this.templateName.equals(templateName)) return false;
		if (this.args != null && !this.args.equals(args)) return false;
		return true;
	}

	/**
	 * Generated from the message recipient address, template name, and args.
	 *
	 * @return
	 */
	@Override
	public int hashCode()
	{
		int result = address != null ? address.hashCode() : 0;
		result = 31 * result + (templateName != null ? templateName.hashCode() : 0);
		result = 31 * result + (args != null ? args.hashCode() : 0);
		return result;
	}

	@Override
	public String toString()
	{
		return new ToStringBuilder(this, ToStringStyle.SIMPLE_STYLE)
				.append("scheduled", scheduled)
				.append("address", address == null? "": address.getAddress())
				.append("templateName", templateName)
				.append("args", args)
				.toString();
	}
}
