/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

package com.portlandwebworks.commons.email.template;

import com.portlandwebworks.commons.html.HtmlUtils;
import org.springframework.beans.factory.BeanNameAware;

import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * This software is the property of Portland Webworks.
 *
 * A rather simplistic templating system that accepts named parameters like ${name:format} where format is optional
 * and replaces the value from the map provided.  This behaves much like a <code>MessageFormat</code> would, with the
 * exception that instead of using a zero-based number to determine which property value to insert, a named property
 * is expected.
 *
 * For example:  given this Map:
 * <map>
 *  <entry>
 *      <key>name</key>
 *      <value>John</value>
 *  </entry>
 *  <entry>
 *      <key>dateEmailed</key>
 *      <value>2009-11-12 00:00:00</value>
 *  </entry>
 * </map>
 *
 * and this pattern:
 *
 * The user named {name} was last sent an email on {dateEmailed,date,MM/dd/yyyy}.
 *
 * The final output would be:
 *
 * The user named John was last sent an email on 11/12/2009.
 *
 *
 * @Author rnorthup
 */
public class SimpleEmailTemplate implements IEmailTemplate, BeanNameAware
{
	// This pattern will match {beanProperty,format,subformat} where the output format is optional.
	private Pattern pattern = Pattern.compile("\\{([\\w\\d\\.\\[\\]_]+)(,([\\d\\w/:\\-\\s#,\\.]+))?\\}");

	private String name;

	private MessageFormat subjectFormat;

	private List<String> orderedSubjectParams = new ArrayList<String>();

	private Object[] subjectValues;

	private MessageFormat bodyFormat;

	private List<String> orderedBodyParams = new ArrayList<String>();

	private Object[] bodyValues;

	Map<String, Object> paramValues;

	private Locale locale;

	private boolean isHtml;

	public SimpleEmailTemplate()
	{
	}

	public SimpleEmailTemplate(String subjectFormat, String bodyFormat)
	{
		this.subjectFormat = _buildMessageFormat(subjectFormat, orderedSubjectParams);
		this.bodyFormat = _buildMessageFormat(bodyFormat, orderedBodyParams);
	}

	private MessageFormat _buildMessageFormat(String format, List<String> orderedArgs)
	{
		StringBuilder finalPattern = new StringBuilder();

		Matcher matcher = pattern.matcher(format);
		int begin = 0;
		int position = 0;
		while (matcher.find()) {
			// First group is the bean name, second group is the pattern with a colon on the front, third match is
			// the pattern by itself.
			String beanProperty = matcher.group(1);

			// Add bean name to list
			orderedArgs.add(beanProperty);

			// Replace bean name in
			finalPattern.append(format.substring(begin, matcher.start(1)));
			finalPattern.append(position++);
			begin = matcher.end(1);
		}

		finalPattern.append(format.substring(begin, format.length()));

		return new MessageFormat(finalPattern.toString());
	}

	private Object[] _buildOrderedParamValues(Map<String, Object> map, List<String> orderedArgs)
	{
		Object[] values = new Object[orderedArgs.size()];

		int position = 0;
		for (String orderedArg : orderedArgs) {
			values[position++] = map.get(orderedArg);
		}

		return values;
	}

	public void setParamValues(Map<String, Object> paramValues)
	{
		this.paramValues = paramValues;

		subjectValues = _buildOrderedParamValues(this.paramValues, orderedSubjectParams);
		bodyValues = _buildOrderedParamValues(this.paramValues, orderedBodyParams);
	}

	public void setName(String name)
	{
		this.name = name;
	}

	protected String getName()
	{
		return name;
	}

	public void setSubjectFormat(String subjectFormat)
	{
		this.subjectFormat = _buildMessageFormat(subjectFormat, orderedSubjectParams);
	}

	public void setBodyFormat(String bodyFormat)
	{
		this.bodyFormat = _buildMessageFormat(bodyFormat, orderedBodyParams);

		isHtml = HtmlUtils.isContentHtml(bodyFormat);
	}

	public String getSubject()
	{
		return subjectFormat.format(subjectValues);
	}

	public String getBody()
	{
		return bodyFormat.format(bodyValues);
	}

	public boolean isHtml()
	{
		return isHtml;
	}

	/**
	 * Allows use of the spring configuration bean ID as the name of this template to allow use of <bean id="name".../>
	 * instead of <bean id="id" name="name".../>.  Note however that this will not work for localized email templates,
	 * which would result in multiple XML nodes with the same ID, only differentiated by locale.
	 * 
	 * @param beanName
	 */
	public void setBeanName(String beanName)
	{
		this.name = beanName;
	}

	public String getLocaleName()
	{
		return locale == null? null: locale.getDisplayName();
	}

	public void setLocaleName(String locale)
	{
		String[] localeParts = locale.split("_");
		if (localeParts.length > 1) {
			this.locale = new Locale(localeParts[0], localeParts[1]);
		}
		else {
			this.locale = new Locale(localeParts[0]);
		}
	}
}
