/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

package com.portlandwebworks.commons.hibernate;

import org.hibernate.*;
import org.hibernate.criterion.Criterion;
import org.hibernate.criterion.DetachedCriteria;
import org.hibernate.criterion.Order;
import org.hibernate.criterion.Projection;
import org.hibernate.transform.ResultTransformer;

import java.util.List;

/**
 * This software is the property of Portland Webworks.
 *
 * This adapter was created to deal with a design flaw in the Hibernate querying API.  DetachedCriteria and Criteria
 * do not share the same interface, so even though everything that a DetachedCriteria does is also done in the Criteria
 * API.  The big difference is that Criteria is associated with a session, which allows it to support methods that
 * execute the underlying query.  A DetachedCriteria object does not have a session and does not support query execution
 * or operations related to query execution.
 *
 * What they really should have done is NOT included methods that execute the query in the Criteria API, so that a Criteria
 * object is always detached from the Session.
 *
 * @Author rnorthup
 */
public class DetachedCriteriaAdapter implements Criteria
{
	private DetachedCriteria detached;

	
	public DetachedCriteriaAdapter(DetachedCriteria detached)
	{
		this.detached = detached;
	}

	public String getAlias()
	{
		return detached.getAlias();
	}

	public Criteria setProjection(Projection projection)
	{
		detached.setProjection(projection);
		return this;
	}

	public Criteria add(Criterion criterion)
	{
		detached.add(criterion);
		return this;
	}

	public Criteria addOrder(Order order)
	{
		detached.addOrder(order);
		return this;
	}

	public Criteria setFetchMode(String path, FetchMode mode) throws HibernateException
	{
		detached.setFetchMode(path, mode);
		return this;
	}

	public Criteria setLockMode(LockMode lockMode)
	{
		detached.setLockMode(lockMode);
		return this;
	}


	public Criteria setLockMode(String alias, LockMode lockMode)
	{
		detached.setLockMode(alias, lockMode);
		return this;
	}


	public Criteria createAlias(String path, String alias) throws HibernateException
	{
		detached.createAlias(path, alias);
		return this;
	}


	public Criteria createAlias(String path, String alias, int joinType) throws HibernateException
	{
		detached.createAlias(path, alias, joinType);
		return this;
	}

	public Criteria createAlias(String s, String s1, int i, Criterion criterion) throws HibernateException
	{
		throw new RuntimeException("Detached query doesn't support this operation");
	}


	public Criteria createCriteria(String path) throws HibernateException
	{
		detached.createCriteria(path);
		return this;
	}


	public Criteria createCriteria(String path, int joinType) throws HibernateException
	{
		detached.createCriteria(path, joinType);
		return this;
	}


	public Criteria createCriteria(String path, String alias) throws HibernateException
	{
		detached.createCriteria(path, alias);
		return this;
	}


	public Criteria createCriteria(String path, String alias, int joinType) throws HibernateException
	{
		detached.createCriteria(path, alias, joinType);
		return this;
	}

	public Criteria createCriteria(String s, String s1, int i, Criterion criterion) throws HibernateException
	{
		return null;  //To change body of implemented methods use File | Settings | File Templates.
	}


	public Criteria setResultTransformer(ResultTransformer resultTransformer)
	{
		detached.setResultTransformer(resultTransformer);
		return this;
	}

	public Criteria setComment(String comment)
	{
		detached.setComment(comment);
		return this;
	}

	public Criteria setMaxResults(int i)
	{
		throw new RuntimeException("Detached queries don't support this operation");
	}

	public Criteria setFirstResult(int i)
	{
		throw new RuntimeException("Detached queries don't support this operation");
	}

	public boolean isReadOnlyInitialized()
	{
		throw new RuntimeException("Detached query doesn't support this operation");
	}

	public boolean isReadOnly()
	{
		throw new RuntimeException("Detached query doesn't support this operation");
	}

	public Criteria setReadOnly(boolean b)
	{
		throw new RuntimeException("Detached query doesn't support this operation");
	}


	public Criteria setFetchSize(int i)
	{
		throw new RuntimeException("Detached queries don't support this operation");
	}

	public Criteria setTimeout(int i)
	{
		throw new RuntimeException("Detached queries don't support this operation");
	}

	public Criteria setCacheable(boolean b)
	{
		throw new RuntimeException("Detached queries don't support this operation");
	}


	public Criteria setCacheRegion(String s)
	{
		throw new RuntimeException("Detached queries don't support this operation");
	}

	public Criteria setFlushMode(FlushMode flushMode)
	{
		throw new RuntimeException("Detached queries don't support this operation");
	}

	public Criteria setCacheMode(CacheMode cacheMode)
	{
		throw new RuntimeException("Detached queries don't support this operation");
	}

	public List list() throws HibernateException
	{
		throw new RuntimeException("Detached queries don't support this operation");
	}

	public ScrollableResults scroll() throws HibernateException
	{
		throw new RuntimeException("Detached queries don't support this operation");
	}

	public ScrollableResults scroll(ScrollMode scrollMode) throws HibernateException
	{
		throw new RuntimeException("Detached queries don't support this operation");
	}

	public Object uniqueResult() throws HibernateException
	{
		throw new RuntimeException("Detached queries don't support this operation");
	}
}
