/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

package com.portlandwebworks.commons.spring.cache;

import net.sf.ehcache.Cache;
import net.sf.ehcache.Element;
import org.aopalliance.intercept.MethodInterceptor;
import org.aopalliance.intercept.MethodInvocation;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Required;

import java.io.Serializable;

/**
 *
 * usage:
 * <code>
 * 
<bean id="methodCacheInterceptor" class="org.taha.interceptor.MethodCacheInterceptor">
  <property name="cache">
    <ref local="methodCache" />
  </property>
</bean>

<bean id="methodCachePointCut" class="org.springframework.aop.support.RegexpMethodPointcutAdvisor">
  <property name="advice">
    <ref local="methodCacheInterceptor"/>
  </property>
  <property name="patterns">
    <list>
      <value>.*methodOne</value>
      <value>.*methodTwo</value>
    </list>
  </property>
</bean>

<bean id="myBean" class="org.springframework.aop.framework.ProxyFactoryBean">
  <property name="target">
   <bean class="org.taha.beans.MyBean"/>
  </property>
  <property name="interceptorNames">
    <list>
      <value>methodCachePointCut</value>
    </list>
  </property>
</bean>

 * </code>
 *
 *
 * @author <a href="mailto:irbouh@gmail.com">Omar Irbouh</a>, modifications by Ronnie Northup
 * @since 2004.10.07
 */
public class MethodCacheInterceptor implements MethodInterceptor
{
	private static final Logger logger = Logger.getLogger(MethodCacheInterceptor.class);

	private Cache cache;

	/**
	 * sets cache name to be used
	 */
	@Required
	public void setCache(Cache cache)
	{
		this.cache = cache;
	}

	/**
	 * main method
	 * caches method result if method is configured for caching
	 * method results must be serializable
	 */
	public Object invoke(MethodInvocation invocation) throws Throwable
	{
		String targetName = invocation.getThis().getClass().getName();
		String methodName = invocation.getMethod().getName();
		Object[] arguments = invocation.getArguments();
		Object result;

		logger.debug("looking for method result in cache");
		String cacheKey = getCacheKey(targetName, methodName, arguments);
		Element element = cache.get(cacheKey);
		if (element == null) {
			//call target/sub-interceptor
			logger.debug("calling intercepted method");
			result = invocation.proceed();

			//cache method result
			logger.debug("caching result");
			element = new Element(cacheKey, (Serializable)result);
			cache.put(element);
		}
		return element.getValue();
	}

	/**
	 * creates cache key: targetName.methodName.argument0.argument1...
	 */
	private String getCacheKey(String targetName, String methodName, Object[] arguments)
	{
		StringBuilder sb = new StringBuilder();
		sb.append(targetName).append(".").append(methodName);
		
		if ((arguments != null) && (arguments.length != 0)) {
			for (Object argument : arguments) {
				sb.append(".").append(argument);
			}
		}

		return sb.toString();
	}
}