/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

package com.portlandwebworks.commons.test;

import org.apache.commons.beanutils.PropertyUtils;
import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.time.DateUtils;
import org.apache.log4j.Logger;

import java.beans.PropertyDescriptor;
import java.lang.reflect.InvocationTargetException;
import java.text.MessageFormat;
import java.util.*;

/**
 * This software is the property of Portland Webworks.
 *
 * Helper class for
 *
 * @Author rnorthup
 */
public class MatchUtils
{
	static Logger logger = Logger.getLogger(MatchUtils.class);

	private static String[] globalIgnoreProperties = { "annotations", "annotationType" };

	/**
	 * Determines if two objects match.
	 * @param left
	 * @param right
	 * @param ignoreProperties
	 * @return
	 */
	public static boolean match(Object left, Object right, String... ignoreProperties)
	{
		return match(left, right, false, ignoreProperties);
	}

	/**
	 * Determines if two objects and all of their aggregate objects match.
	 * 
	 * @param left
	 * @param right
	 * @param ignoreProperties
	 * @return
	 */
	public static boolean deepMatch(Object left, Object right, String... ignoreProperties)
	{
		return match(left, right, true, ignoreProperties);
	}

	private static boolean match(Object left, Object right, boolean recurse, String... ignoreProperties)
	{
		PropertyDescriptor[] tmp = PropertyUtils.getPropertyDescriptors(left);
		try {
			for (PropertyDescriptor descriptor : tmp) {
				if (ignoreProperties != null && ArrayUtils.contains(ignoreProperties, descriptor.getName())) {
					if (logger.isDebugEnabled())
						logger.debug(MessageFormat.format("ignore: {0}.{1}", left.getClass(), descriptor.getName()));
					continue;
				}

/*

				if (ArrayUtils.contains(globalIgnoreProperties, descriptor.getName())) {
					if (logger.isDebugEnabled())
						logger.debug(MessageFormat.format("global ignore: {0}.{1}", left.getClass(), descriptor.getName()));
					continue;
				}
*/

				boolean match = false;
			
				Object leftVal = descriptor.getReadMethod().invoke(left);
				Object rightVal = descriptor.getReadMethod().invoke(right);

				if (leftVal == null) {
					match = rightVal == null;
				}
				else if (rightVal == null) {
					match = false;
				}
				else {
					Class<?> propertyType = descriptor.getPropertyType();
					if (Collection.class.isAssignableFrom(propertyType)) {
						ArrayList leftc = new ArrayList((Collection)leftVal);
						LinkedList rightc = new LinkedList((Collection)rightVal);

						// If the number of elements in both collections do not match, fail.
						if (leftc.size() != rightc.size()) {
							match = false;
						}
						else {
							// Iterate through the LHS collection, trying to find a match in the RHS collection.
							for (Object lefto : leftc) {
								for (int i = 0; i < rightc.size(); i++) {
									Object righto = rightc.get(i);
									// Recurse on the nested object.
									//TODO: send along a list of nested ignore properties by finding all ignore props
									//TODO: that have dot notation, and stripping off the value up to the first dot.
									match = match(lefto, righto, recurse, ignoreProperties);
									if (match) {
										// Eliminate this RHS value for any future matches.
										rightc.remove(i);
										break;
									}
								}
								if (! match) {
									// No reason to continue evaluating LHS objects.
									break;
								}
							}
							// Every LHS object has a match in the RHS collection.
							match = true;
						}
					}
					else if (Class.class.isAssignableFrom(propertyType) ||
							Number.class.isAssignableFrom(propertyType) ||
							propertyType.isPrimitive() ||
							propertyType.equals(Byte.class) ||
							propertyType.equals(Character.class) ||
							propertyType.equals(String.class) ||
							propertyType.isEnum() ||
							propertyType.equals(Boolean.class)) {
						// Wrapper classes for primitive types, enums and Strings can use equals()
						match = (leftVal).equals(rightVal);
					}
					else if (propertyType.equals(Date.class)) {
						// This special handling allows a java.util.Date to be compared to a java.sql.Timestamp, which is
						// what Hibernate likes to stick in date fields when transforming query results.
						match = DateUtils.truncate((Date)leftVal, Calendar.MILLISECOND).equals(
								DateUtils.truncate((Date)rightVal, Calendar.MILLISECOND));
					}
					else {
						// Recurse on the nested object.
						//TODO: send along a list of nested ignore properties by finding all ignore props
						//TODO: that have dot notation, and stripping off the value up to the first dot.
						match = match(leftVal, rightVal, recurse, ignoreProperties);
					}
				}

				if (! match) {

					if (logger.isDebugEnabled())
						logger.debug(MessageFormat.format("{0}.{1}: {2} != {3}",
								left.getClass(), descriptor.getName(), leftVal, rightVal));

					return false;
				}
			}
			return true;
		}
		catch (IllegalAccessException e) {
			return false;
		}
		catch (InvocationTargetException e) {
			return false;
		}
	}

}
